@@ -1,6 +1,8 @@
 """Londiste setup and sanity checker.
 """
 
+from typing import Optional, Sequence
+
 import os
 import re
 import sys
@@ -25,7 +27,13 @@ class LondisteSetup(CascadeAdmin):
     provider_location = None
 
     commands_without_pidfile = CascadeAdmin.commands_without_pidfile + [
-        'tables', 'seqs', 'missing', 'show-handlers']
+        'tables', 'seqs', 'missing', 'show-handlers', 'wait-sync',
+    ]
+
+    register_only_tables: Optional[Sequence[str]] = None
+    register_only_seqs: Optional[Sequence[str]] = None
+    register_skip_tables: Optional[Sequence[str]] = None
+    register_skip_seqs: Optional[Sequence[str]] = None
 
     def install_code(self, db):
         self.extra_objs = [
@@ -48,6 +56,11 @@ class LondisteSetup(CascadeAdmin):
 
         self.lock_timeout = self.cf.getfloat('lock_timeout', 10)
 
+        self.register_only_tables = self.cf.getlist("register_only_tables", [])
+        self.register_only_seqs = self.cf.getlist("register_only_seqs", [])
+        self.register_skip_tables = self.cf.getlist("register_skip_tables", [])
+        self.register_skip_seqs = self.cf.getlist("register_skip_seqs", [])
+
         load_handler_modules(self.cf)
 
     def init_optparse(self, parser=None):
@@ -108,6 +121,10 @@ class LondisteSetup(CascadeAdmin):
         pcurs.execute(q, [self.set_name])
         for row in pcurs.fetchall():
             tbl = row['table_name']
+            if self.register_only_tables and tbl not in self.register_only_tables:
+                continue
+            if self.register_skip_tables and tbl in self.register_skip_tables:
+                continue
             q = "select * from londiste.global_add_table(%s, %s)"
             ncurs.execute(q, [self.set_name, tbl])
 
@@ -117,6 +134,10 @@ class LondisteSetup(CascadeAdmin):
         for row in pcurs.fetchall():
             seq = row['seq_name']
             val = row['last_value']
+            if self.register_only_seqs and seq not in self.register_only_seqs:
+                continue
+            if self.register_skip_seqs and seq in self.register_skip_seqs:
+                continue
             q = "select * from londiste.global_update_seq(%s, %s, %s)"
             ncurs.execute(q, [self.set_name, seq, val])
 
@@ -168,7 +189,7 @@ class LondisteSetup(CascadeAdmin):
         # search for usable copy node if requested & needed
         if (self.options.find_copy_node and create_flags != 0
                 and needs_tbl and not self.is_root()):
-            src_name, _, _ = find_copy_source(self, self.queue_name, args, None, self.provider_location)
+            src_name, _, _ = find_copy_source(self, self.queue_name, args, "?", self.provider_location)
             self.options.copy_node = src_name
             self.close_database('provider_db')
             src_db = self.get_provider_db()
@@ -310,6 +331,10 @@ class LondisteSetup(CascadeAdmin):
 
     def sync_table_list(self, dst_curs, src_tbls, dst_tbls):
         for tbl in src_tbls.keys():
+            if self.register_only_tables and tbl not in self.register_only_tables:
+                continue
+            if self.register_skip_tables and tbl in self.register_skip_tables:
+                continue
             q = "select * from londiste.global_add_table(%s, %s)"
             if tbl not in dst_tbls:
                 self.log.info("Table %s info missing from subscriber, adding", tbl)
@@ -448,6 +473,10 @@ class LondisteSetup(CascadeAdmin):
     def sync_seq_list(self, dst_curs, src_seqs, dst_seqs):
         for seq in src_seqs.keys():
             q = "select * from londiste.global_update_seq(%s, %s, %s)"
+            if self.register_only_seqs and seq not in self.register_only_seqs:
+                continue
+            if self.register_skip_seqs and seq in self.register_skip_seqs:
+                continue
             if seq not in dst_seqs:
                 self.log.info("Sequence %s info missing from subscriber, adding", seq)
                 self.exec_cmd(dst_curs, q, [self.set_name, seq, src_seqs[seq]['last_value']])
@@ -586,7 +615,8 @@ class LondisteSetup(CascadeAdmin):
 
         for fn in files:
             fname = os.path.basename(fn)
-            sql = open(fn, "r").read()
+            with open(fn, "r", encoding="utf8") as f:
+                sql = f.read()
             attrs = ExecAttrs(sql=sql)
             q = "select * from londiste.execute_start(%s, %s, %s, true, %s)"
             res = self.exec_cmd(db, q, [self.queue_name, fname, sql, attrs.to_urlenc()], commit=False)
@@ -682,7 +712,7 @@ class LondisteSetup(CascadeAdmin):
                 rc = re.compile(glob2regex(a))
                 for x in full_list:
                     if rc.match(x):
-                        if not x in res_map:
+                        if x not in res_map:
                             res_map[x] = 1
                             res_list.append(x)
             else:
