@@ -4,13 +4,18 @@
 Per-table decision how to create trigger, copy data and apply events.
 """
 
-from __future__ import division, absolute_import, print_function
+from typing import List, Dict, Any, Sequence, Tuple, Optional
 
-import sys
 import json
 import logging
+import sys
+
 import skytools
+from skytools.basetypes import Cursor
+
 import londiste.handlers
+import londiste.util
+
 
 _ = """
 
@@ -41,8 +46,14 @@ plain londiste:
 __all__ = ['RowCache', 'BaseHandler', 'build_handler', 'EncodingValidator',
            'load_handler_modules', 'create_handler_string']
 
-class RowCache(object):
-    def __init__(self, table_name):
+
+class RowCache:
+
+    table_name: str
+    keys: Dict[str, str]
+    rows: List[Tuple[Any, ...]]
+
+    def __init__(self, table_name: str) -> None:
         self.table_name = table_name
         self.keys = {}
         self.rows = []
@@ -61,7 +72,7 @@ class RowCache(object):
 
     def get_fields(self):
         row = [None] * len(self.keys)
-        for k, i in self.keys.keys():
+        for k, i in self.keys.items():
             row[i] = k
         return tuple(row)
 
@@ -69,7 +80,8 @@ class RowCache(object):
         fields = self.get_fields()
         skytools.magic_insert(curs, self.table_name, self.rows, fields)
 
-class BaseHandler(object):
+
+class BaseHandler:
     """Defines base API, does nothing.
     """
     handler_name = 'nop'
@@ -161,7 +173,7 @@ class BaseHandler(object):
         """ Use if you want to filter data """
         return ''
 
-    def real_copy(self, src_tablename, src_curs, dst_curs, column_list):
+    def real_copy(self, src_tablename: str, src_curs: Cursor, dst_curs: Cursor, column_list: List[str]):
         """do actual table copy and return tuple with number of bytes and rows
         copied
         """
@@ -170,6 +182,30 @@ class BaseHandler(object):
                                   column_list, condition,
                                   dst_tablename=self.dest_table)
 
+    def real_copy_threaded(
+        self,
+        src_real_table: str,
+        src_curs: Cursor,
+        dst_db_connstr: str,
+        column_list: Sequence[str],
+        config_file: str,
+        config_section: str,
+        parallel: int = 1,
+    ) -> Tuple[int, int]:
+        with skytools.connect_database(dst_db_connstr) as dst_db:
+            with dst_db.cursor() as dst_curs:
+                condition = self.get_copy_condition(src_curs, dst_curs)
+            dst_db.commit()
+
+        return londiste.util.full_copy_parallel(
+            src_real_table, src_curs,
+            dst_db_connstr=dst_db_connstr,
+            dst_tablename=self.dest_table,
+            column_list=column_list,
+            condition=condition,
+            parallel=parallel,
+        )
+
     def needs_table(self):
         """Does the handler need the table to exist on destination."""
         return True
@@ -179,6 +215,10 @@ class BaseHandler(object):
         """Load conf."""
         pass
 
+    def get_copy_event(self, ev, queue_name):
+        """Get event copy for destination queue."""
+        return ev
+
 
 class TableHandler(BaseHandler):
     """Default Londiste handler, inserts events into tables with plain SQL.
@@ -199,7 +239,7 @@ class TableHandler(BaseHandler):
     allow_sql_event = 1
 
     def __init__(self, table_name, args, dest_table):
-        super(TableHandler, self).__init__(table_name, args, dest_table)
+        super().__init__(table_name, args, dest_table)
 
         enc = args.get('encoding')
         if enc:
@@ -208,7 +248,7 @@ class TableHandler(BaseHandler):
             self.encoding_validator = None
 
     def get_config(self):
-        conf = super(TableHandler, self).get_config()
+        conf = super().get_config()
         conf.ignore_truncate = self.get_arg('ignore_truncate', [0, 1], 0)
         return conf
 
@@ -276,13 +316,45 @@ class TableHandler(BaseHandler):
                                   dst_tablename=self.dest_table,
                                   write_hook=_write_hook)
 
+    def real_copy_threaded(
+        self,
+        src_real_table: str,
+        src_curs: Cursor,
+        dst_db_connstr: str,
+        column_list: Sequence[str],
+        config_file: str,
+        config_section: str,
+        parallel: int = 1,
+    ) -> Tuple[int, int]:
+        with skytools.connect_database(dst_db_connstr) as dst_db:
+            with dst_db.cursor() as dst_curs:
+                condition = self.get_copy_condition(src_curs, dst_curs)
+            dst_db.commit()
+
+        _write_hook: Optional[londiste.util.WriteHook]
+        if self.encoding_validator:
+            def _write_hook(obj, data):
+                return self.encoding_validator.validate_copy(data, column_list, src_real_table)
+        else:
+            _write_hook = None
+
+        return londiste.util.full_copy_parallel(
+            src_real_table, src_curs,
+            dst_db_connstr=dst_db_connstr,
+            column_list=column_list,
+            condition=condition,
+            dst_tablename=self.dest_table,
+            parallel=parallel,
+            write_hook=_write_hook,
+        )
+
 
 #------------------------------------------------------------------------------
 # ENCODING VALIDATOR
 #------------------------------------------------------------------------------
 
-class EncodingValidator(object):
-    def __init__(self, log, encoding='utf-8', replacement=u'\ufffd'):
+class EncodingValidator:
+    def __init__(self, log, encoding='utf-8', replacement='\ufffd'):
         """validates the correctness of given encoding. when data contains
         illegal symbols, replaces them with <replacement> and logs the
         incident
@@ -334,7 +406,7 @@ class EncodingValidator(object):
         ok, u = skytools.safe_utf8_decode(value)
         if ok:
             return value
-        _pfx = pfx and (pfx+': ') or ""
+        _pfx = pfx and (pfx + ': ') or ""
         self.log.info('%sFixed invalid UTF8 in string <%s>', _pfx, value)
         return u.encode('utf8')
 
@@ -342,10 +414,12 @@ class EncodingValidator(object):
 # handler management
 #
 
+
 _handler_map = {
     'londiste': TableHandler,
 }
 
+
 def register_handler_module(modname, cf):
     """Import and module and register handlers."""
     try:
@@ -354,10 +428,11 @@ def register_handler_module(modname, cf):
         print("Failed to load handler module: %s" % (modname,))
         return
     m = sys.modules[modname]
-    for h in m.__londiste_handlers__:
+    for h in getattr(m, "__londiste_handlers__"):
         h.load_conf(cf)
         _handler_map[h.handler_name] = h
 
+
 def _parse_arglist(arglist):
     args = {}
     for arg in arglist or []:
@@ -368,6 +443,7 @@ def _parse_arglist(arglist):
         args[key] = val.strip()
     return args
 
+
 def create_handler_string(name, arglist):
     handler = name
     if name.find('(') >= 0:
@@ -378,21 +454,23 @@ def create_handler_string(name, arglist):
         handler = '%s(%s)' % (handler, astr)
     return handler
 
+
 def _parse_handler(hstr):
     """Parse result of create_handler_string()."""
     args = {}
     name = hstr
     pos = hstr.find('(')
     if pos > 0:
-        name = hstr[ : pos]
+        name = hstr[: pos]
         if hstr[-1] != ')':
             raise Exception('invalid handler format: %s' % hstr)
-        astr = hstr[pos + 1 : -1]
+        astr = hstr[pos + 1: -1]
         if astr:
             astr = astr.replace(',', '&')
             args = skytools.db_urldecode(astr)
     return (name, args)
 
+
 def build_handler(tblname, hstr, dest_table=None):
     """Parse and initialize handler.
 
@@ -405,6 +483,7 @@ def build_handler(tblname, hstr, dest_table=None):
         dest_table = tblname
     return klass(tblname, args, dest_table)
 
+
 def load_handler_modules(cf):
     """Load and register modules from config."""
     lst = londiste.handlers.DEFAULT_HANDLERS
@@ -413,6 +492,7 @@ def load_handler_modules(cf):
     for m in lst:
         register_handler_module(m, cf)
 
+
 def show(mods):
     if not mods:
         for n in _handler_map:
@@ -428,3 +508,4 @@ def show(mods):
             if desc:
                 desc = desc.strip()
             print("%s - %s" % (n, desc))
+
