@@ -4,7 +4,15 @@ set -e
 set -x
 
 tests="qtable"
-tests="simple qtable obfuscate shard register"
+tests="
+    simple
+    qtable
+    obfuscate
+    shard
+    register
+    takeover-merge
+"
+tests="shard"
 
 PAGER=cat
 export PAGER
